@@ -6,4 +6,6 @@ from tethys_utils import (
     s3,
     data_io,
     titan,
+    backwards,
+    raster
 )
