@@ -17,7 +17,7 @@ import tethys_data_models as tdm
 from tethysts.utils import get_object_s3, s3_client, read_json_zstd, read_pkl_zstd, chunk_filters
 from tethys_utils.misc import make_run_date_key, write_json_zstd, path_date_parser
 # from misc import make_run_date_key, write_json_zstd, write_pkl_zstd
-from tethys_utils.processing import compare_xrs, process_station_summ, stats_for_dataset_metadata
+from tethys_utils.processing import stats_for_dataset_metadata
 # from processing import compare_xrs, process_station_summ, stats_for_dataset_metadata
 from datetime import date, datetime
 from tethysts import Tethys
@@ -26,6 +26,7 @@ from time import sleep
 import pathlib
 import concurrent.futures
 import multiprocessing as mp
+import copy
 
 ############################################
 ### Parameters
@@ -73,7 +74,7 @@ def put_file_s3(s3, bucket, key, file_path, metadata, content_type, retries=5):
     return obj_out
 
 
-def copy_object_s3(s3, source_bucket, source_key, dest_bucket, dest_key, retries=5):
+def copy_object_s3(s3, source_bucket, dest_bucket, source_key, dest_key, retries=5):
     """
     Copies an object in an S3 database to another location in an S3 database. They must have the same fundemental connection_config. All metadata is copied to the new object.
     https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.copy_object
@@ -83,7 +84,7 @@ def copy_object_s3(s3, source_bucket, source_key, dest_bucket, dest_key, retries
     counter = retries
     while counter > 0:
         try:
-            obj2 = s3.copy_object(Bucket=dest_bucket, Key=dest_key, CopySource=source_dict, MetadataDirective='COPY')
+            resp = s3.copy_object(Bucket=dest_bucket, Key=dest_key, CopySource=source_dict, MetadataDirective='COPY')
             break
         except bc_exceptions.ConnectionClosedError as err:
             print(err)
@@ -100,7 +101,28 @@ def copy_object_s3(s3, source_bucket, source_key, dest_bucket, dest_key, retries
             print('...trying again...')
             sleep(5)
 
-    return obj2
+    return resp
+
+
+def multi_copy_object_s3(s3, source_bucket, dest_bucket, source_dest_keys, retries=5, threads=30):
+    """
+    Same as the copy_object_s3 except with multi threading. the input source_dest_keys must be a list of dictionaries with keys named source_key and dest_key.
+    """
+    keys = copy.deepcopy(source_dest_keys)
+
+    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+        futures = []
+        for key in keys:
+            key['s3'] = s3
+            key['source_bucket'] = source_bucket
+            key['dest_bucket'] = dest_bucket
+            f = executor.submit(copy_object_s3, **key)
+            futures.append(f)
+        runs = concurrent.futures.wait(futures)
+
+    resp_list = [r.result() for r in runs[0]]
+
+    return resp_list
 
 
 def put_interim_results_s3(s3, bucket, file_path, run_id, system_version=4, retries=5):
@@ -333,29 +355,6 @@ def get_remote_results_chunks(bucket, s3=None, connection_config=None, public_ur
     return rem_ds
 
 
-
-# def put_remote_station(s3, bucket, station, run_date=None, system_version=4):
-#     """
-
-#     """
-#     run_date_key = make_run_date_key(run_date)
-
-#     dataset_id = station['dataset_id']
-#     station_id = station['station_id']
-
-#     stn4 = Station(**station)
-
-#     stn5 = orjson.loads(stn4.json(exclude_none=True))
-
-#     stn_obj = write_json_zstd(stn5)
-
-#     stn_key = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=dataset_id, station_id=station_id)
-
-#     _ = put_object_s3(s3, bucket, stn_key, stn_obj, {'run_date': run_date_key}, 'application/json')
-
-#     return stn5
-
-
 def put_remote_agg_stations(dataset_id, bucket, s3, public_url=None, threads=60, system_version=4):
     """
 
@@ -447,208 +446,6 @@ def put_remote_agg_datasets(bucket, s3, public_url=None, threads=30, system_vers
     return ds_list
 
 
-def compare_datasets_from_s3(connection_config, bucket, new_data, add_old=False, last_run_date_key=None, public_url=None, version=3):
-    """
-    Parameters
-    ----------
-    connection_config : dict
-        A dictionary of the connection info necessary to establish an S3 connection.
-    bucket : str
-        The S3 bucket.
-    new_data : xr.Dataset
-        The new data that should be compared to existing data in S3.
-    add_old : bool
-        Should the data in the S3 be added to the output?
-    last_run_date_key : str
-        Specify the last run key instead of having the function figure it out. The function will do a check to make sure that the key exists.
-    public_url : str
-        Optional if there is a public URL to the object instead of using the S3 API directly.
-
-    Returns
-    -------
-    xr.Dataset
-        Of the data that should be updated.
-    """
-    ## Determine the parameter, station_id, and dataset_id
-    vars1 = list(new_data.variables)
-    dataset = [new_data[v].attrs for v in vars1 if 'dataset_id' in new_data[v].attrs][0]
-    dataset_id = dataset['dataset_id']
-    # result_type = dataset['result_type']
-    station_id = str(new_data.squeeze()['station_id'].values)
-
-    key_dict = {'dataset_id': dataset_id, 'station_id': station_id}
-
-    base_key_pattern = tdm.utils.key_patterns[version]['results']
-
-    ## Get list of keys
-    if isinstance(last_run_date_key, str):
-        key_dict.update({'run_date': last_run_date_key})
-        last_key = base_key_pattern.format(**key_dict)
-    else:
-        last_key = None
-
-    ## Get previous data and compare
-    if isinstance(public_url, str):
-        conn_config = public_url
-    else:
-        conn_config = connection_config
-
-    if isinstance(last_key, str):
-        # last_key = last_key1.iloc[0]['Key']
-        p_old_one = get_object_s3(last_key, conn_config, bucket, 'zstd')
-        xr_old_one = xr.load_dataset(p_old_one)
-
-        up1 = compare_xrs(xr_old_one, new_data, add_old=add_old)
-    else:
-        # print('No prior data found in S3. All data will be returned.')
-        up1 = new_data.copy()
-
-    return up1
-
-
-# def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads=30, public_url=None, version=3):
-#     """
-#     Parameters
-#     ----------
-#     processing_code : int
-#         The processing code to determine how the input data should be processed.
-#     data_dict : dict of lists
-#         A dictionary with the keys as the dataset_ids and teh values as lists of zstd compressed xr.Datasets.
-#     remote : dict
-#         Dict of a connection_config and bucket:
-#         conn_config : dict
-#             A dictionary of the connection info necessary to establish an S3 connection.
-#         bucket : str
-#             The S3 bucket.
-#     threads : int
-#         The number of threads to use to process the data.
-#     public_url : str
-#         Optional if there is a public URL to the object instead of using the S3 API directly.
-
-#     Returns
-#     -------
-#     None
-
-#     """
-#     ### Parameters
-#     conn_config = remote['connection_config']
-#     bucket = remote['bucket']
-
-#     if processing_code in [2, 3, 6]:
-#         add_old = True
-#     elif processing_code in [1, 4, 5]:
-#         add_old = False
-#     else:
-#         raise ValueError('processing_code does not exist.')
-
-#     if processing_code in [4, 5]:
-#         no_comparison = True
-#     else:
-#         no_comparison = False
-
-#     ### Run update
-
-#     s3 = s3_client(conn_config, threads)
-
-#     for ds_id, results in data_dict.items():
-#         print('--dataset_id: ' + ds_id)
-
-#         if len(results) > 0:
-
-#             run_date_key = run_date_dict[ds_id]
-
-#             # Create the Key info dict
-#             prefix = tdm.utils.key_patterns[version]['results'].split('{station_id}')[0].format(dataset_id=ds_id)
-
-#             keys1 = list_objects_s3(s3, bucket, prefix)
-#             obj_df1 = keys1[keys1.Key.str.contains('results.nc')].copy()
-#             # obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
-#             obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])
-
-#             last_date1 = obj_df1.groupby(['station_id'])['KeyDate'].last().reset_index()
-
-
-#             def update_result(result):
-#                 """
-
-#                 """
-#                 ## Process data
-#                 try:
-#                     new1 = xr.load_dataset(read_pkl_zstd(result, False))
-#                 except:
-#                     print('Data could not be opened')
-#                     return None
-
-#                 stn_id = str(new1.squeeze()['station_id'].values)
-#                 # print('station_id: ' + stn_id)
-
-#                 vars1 = list(new1.variables)
-#                 parameter = [v for v in vars1 if 'dataset_id' in new1[v].attrs][0]
-#                 result_attrs = new1[parameter].attrs.copy()
-
-#                 ds_id = result_attrs['dataset_id']
-#                 mod_date_key = new1.attrs['history'].split(':')[0]
-
-#                 if no_comparison:
-#                     up1 = new1
-#                 else:
-#                     last_date_key_df = last_date1[last_date1['station_id'] == stn_id]
-#                     if last_date_key_df.empty:
-#                         last_date_key = None
-#                     else:
-#                         last_date_key = make_run_date_key(last_date_key_df['KeyDate'].iloc[0])
-
-#                     up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, last_run_date_key=last_date_key, public_url=public_url, version=version)
-
-#                 ## Save results
-#                 if isinstance(up1, xr.Dataset) and (len(up1[parameter].time) > 0):
-
-#                     # print('Save results')
-#                     key_dict = {'dataset_id': ds_id, 'station_id': stn_id, 'run_date': run_date_key}
-
-#                     new_key = tdm.utils.key_patterns[version]['results'].format(**key_dict)
-
-#                     cctx = zstd.ZstdCompressor(level=1)
-#                     c_obj = cctx.compress(up1.to_netcdf())
-
-#                     obj_resp = put_object_s3(s3, bucket, new_key, c_obj, {'run_date': mod_date_key}, 'application/zstd')
-
-#                     ## Process stn data
-#                     # print('Save station data')
-
-#                     ## Process object key infos
-#                     stn_obj_df1 = obj_df1[obj_df1['station_id'] == stn_id].drop(['LastModified', 'station_id'], axis=1).copy()
-#                     keys = stn_obj_df1['Key'].unique()
-#                     new_key_info = [new_key, obj_resp['ResponseMetadata']['HTTPHeaders']['etag'].replace('"', ''), len(c_obj), pd.Timestamp(run_date_key).tz_localize(None)]
-
-#                     if new_key in keys:
-#                         stn_obj_df1[stn_obj_df1['Key'] == new_key] = new_key_info
-#                     else:
-#                         stn_obj_df1.loc[len(stn_obj_df1)] = new_key_info
-
-#                     info1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate']) for i, row in stn_obj_df1.iterrows()]
-
-#                     ## Final station processing and saving
-#                     stn_m = process_station_summ(up1, info1, mod_date=mod_date_key)
-
-#                     up_stns = put_remote_station(s3, bucket, stn_m, run_date=mod_date_key, version=version)
-
-#                 else:
-#                     print('No new data to update')
-
-#                 ## Get rid of big objects
-#                 new1 = None
-#                 up1 = None
-
-
-#             ## Run the threadpool
-#             # output = ThreadPool(threads).imap_unordered(update_result, results)
-#             with ThreadPool(threads) as pool:
-#                 output = pool.map(update_result, results)
-#                 pool.close()
-#                 pool.join()
-
-
 def update_versions(version_data, bucket, connection_config, public_url=None, system_version=4):
     """
 
@@ -669,12 +466,18 @@ def update_versions(version_data, bucket, connection_config, public_url=None, sy
     if v_obj1 is not None:
         v2 = read_json_zstd(v_obj1)
 
-        version_list = []
-        for v in v2:
-            if version_dict1['version_date'] == v['version_date']:
-                version_list.append(version_dict1)
-            else:
-                version_list.append(v)
+        exists = any([version_dict1['version_date'].isoformat() == v['version_date'] for v in v2])
+
+        if exists:
+            version_list = []
+            for v in v2:
+                if version_dict1['version_date'].isoformat() == v['version_date']:
+                    version_list.append(version_dict1)
+                else:
+                    version_list.append(v)
+        else:
+            version_list = v2.copy()
+            version_list.append(version_dict1)
     else:
         version_list = [version_dict1]
 
@@ -689,16 +492,59 @@ def update_versions(version_data, bucket, connection_config, public_url=None, sy
     return version_list
 
 
+# def update_station_s3(stn_data, bucket, connection_config, public_url=None, system_version=4):
+#     """
+
+#     """
+#     ## Check version_dict data model
+#     stn_dict1 = tdm.dataset.Station(**stn_data).dict(exclude_none=True)
+
+#     ## Initial set up
+#     s3 = s3_client(connection_config)
+
+#     dataset_id = stn_dict1['dataset_id']
+#     stn_id = stn_dict1['station_id']
+
+#     ## Get the versions
+#     stn_key = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=dataset_id, station_id=stn_id)
+
+#     stn_obj1 = get_object_s3(stn_key, bucket, s3=s3, public_url=public_url, counter=1)
+
+#     if stn_obj1 is not None:
+#         s2 = read_json_zstd(stn_obj1)
+
+#         stn_list = []
+#         for s in s2:
+#             if stn_dict1['version_date'] == v['version_date']:
+#                 version_list.append(version_dict1)
+#             else:
+#                 version_list.append(v)
+#     else:
+#         stn_dict2 = stn_dict1
+
+#     ## Save files
+#     run_date_key = make_run_date_key()
+
+#     # versions
+#     stn_obj = write_json_zstd(stn_dict2)
+
+#     _ = put_object_s3(s3, bucket, stn_key, stn_obj, {'run_date': run_date_key}, 'application/json')
+
+#     return stn_dict2
+
+
 def put_result(s3, bucket, results_path, system_version=4):
     """
 
     """
     file_name = os.path.split(results_path)[1]
-    ds_id, version_date_key, stn_id, chunk_id, hash1, _ = file_name.split('_')
+    ds_id, version_date_key, stn_id, chunk_id, _, _ = file_name.split('_')
 
     key_name = tdm.utils.key_patterns[system_version]['results'].format(dataset_id=ds_id, version_date=version_date_key, station_id=stn_id, chunk_id=chunk_id)
 
-    metadata = {'chunk_hash': hash1}
+    run_date_key = make_run_date_key()
+
+    metadata = {'run_date': run_date_key}
 
     _ = put_file_s3(s3, bucket, key_name, results_path, metadata, 'application/zstd', retries=5)
 
@@ -714,156 +560,87 @@ def put_station(s3, bucket, stn_path, system_version=4):
 
     key_name = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=ds_id, station_id=stn_id)
 
-    metadata = {}
+    run_date_key = make_run_date_key()
+
+    metadata = {'run_date': run_date_key}
 
     _ = put_file_s3(s3, bucket, key_name, stn_path, metadata, 'application/zstd', retries=5)
 
     return key_name
 
 
-def delete_result_objects_s3(conn_config, bucket, dataset_ids=None, keep_last=10, threads=50, version=3):
-    """
-    Function to delete Tethys result objects including all object versions.
-
-    Parameters
-    ----------
-    conn_config : dict
-        A dictionary of the connection info necessary to establish an S3 connection.
-    bucket : str
-        The s3 bucket.
-    dataset_ids : list, str, or None
-        The specific datasets that should have the results objects removed. None will remove results objects from all datasets in the bucket.
-    keep_last : int
-        That last number of runs that should be kept. E.g. a value of 4 will kepp the last 4 runs and remove all prior runs.
-    threads : int
-        The number of concurrent threads to use when aggregating the stations and datasets.
-
-    Returns
-    -------
-    list of keys deleted
-    """
-    s3 = s3_client(conn_config, threads)
-
-    if isinstance(dataset_ids, str):
-        dataset_ids = [dataset_ids]
-
-    prefix = tdm.utils.key_patterns[version]['results'].split('{dataset_id}')[0]
-
-    obj_list = list_object_versions_s3(s3, bucket, prefix)
-
-    obj_list1 = obj_list[obj_list.KeyDate.notnull()].copy()
-    key_split = obj_list1['Key'].str.split('/')
-    obj_list1['dataset_id'] = key_split.apply(lambda x: x[2])
-    obj_list1['station_id'] = key_split.apply(lambda x: x[3])
-
-    if isinstance(dataset_ids, list):
-        obj_list1 = obj_list1[obj_list1['dataset_id'].isin(dataset_ids)].copy()
-
-    ## Get the keys to the objects that should be removed
-    obj_list2 = obj_list1.groupby(['dataset_id', 'station_id'])
-
-    rem_keys = []
-    for i, row in obj_list2:
-        rem1 = row.sort_values('KeyDate', ascending=False).iloc[keep_last:]
-        for i2, row2 in rem1.iterrows():
-            rem_keys.extend([{'Key': row2['Key'], 'VersionId': row2['VersionId']}])
-
-    if len(rem_keys) > 0:
-        ## Split them into 1000 key chunks
-        rem_keys_chunks = np.array_split(rem_keys, int(np.ceil(len(rem_keys)/1000)))
-
-        ## Run through and delete the objects...
-        for keys in rem_keys_chunks:
-            resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})
-
-    print(str(len(rem_keys)) + ' objects removed')
-    print('Updating stations and dataset json files...')
-
-    dataset_list = obj_list1['dataset_id'].unique().tolist()
-
-    for ds in dataset_list:
-        ds_stations = put_remote_agg_stations(s3, bucket, ds, threads)
-
-    ### Aggregate all datasets for the bucket
-    ds_all = put_remote_agg_datasets(s3, bucket, threads)
-
-    return rem_keys
-
-
-# def process_results_versions(dataset_list, connection_config, bucket, version=4, version_date=None, version_data=None):
+# def delete_result_objects_s3(conn_config, bucket, dataset_ids=None, keep_last=10, threads=50, version=3):
 #     """
-#     Function to process the run date keys for all datasets for the extraction. These are specific to each processing_code.
+#     Function to delete Tethys result objects including all object versions.
 
 #     Parameters
 #     ----------
-#     processing_code : int
-#         The processing code to determine how the input data should be processed.
-#     dataset_list : list of dict
-#         The list of datasets, which is the output of the process_datasets function.
-#     remote : dict
-#         Dict of a connection_config and bucket:
-#         conn_config : dict
-#             A dictionary of the connection info necessary to establish an S3 connection.
-#         bucket : str
-#             The S3 bucket.
-#     run_date : str, datetime, date, pd.Timestamp, or None
-#         The run_date to use for processing. If None, then the run_date will be generated.
-#     save_interval_hours : int
-#         The frequency by which the datasets should have run dates saved as in Tethys. Essentially the question is, how frequently do you want to keep a record of the data changes? This value is in hours and the default is 2 weeks. This does not effect datasets with processing_code 3 as the run_date always stays the same. A processing_code of 4 oe 5 always creates a new run_date.
+#     conn_config : dict
+#         A dictionary of the connection info necessary to establish an S3 connection.
+#     bucket : str
+#         The s3 bucket.
+#     dataset_ids : list, str, or None
+#         The specific datasets that should have the results objects removed. None will remove results objects from all datasets in the bucket.
+#     keep_last : int
+#         That last number of runs that should be kept. E.g. a value of 4 will kepp the last 4 runs and remove all prior runs.
+#     threads : int
+#         The number of concurrent threads to use when aggregating the stations and datasets.
 
 #     Returns
 #     -------
-#     run_date_dict : dict of str
+#     list of keys deleted
 #     """
-#     if isinstance(version_date, (str, datetime, date, pd.Timestamp)):
-#         run_date1 = pd.Timestamp(version_date)
-#     else:
-#         run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None).round('s')
+#     s3 = s3_client(conn_config, threads)
+
+#     if isinstance(dataset_ids, str):
+#         dataset_ids = [dataset_ids]
+
+#     prefix = tdm.utils.key_patterns[version]['results'].split('{dataset_id}')[0]
+
+#     obj_list = list_object_versions_s3(s3, bucket, prefix)
+
+#     obj_list1 = obj_list[obj_list.KeyDate.notnull()].copy()
+#     key_split = obj_list1['Key'].str.split('/')
+#     obj_list1['dataset_id'] = key_split.apply(lambda x: x[2])
+#     obj_list1['station_id'] = key_split.apply(lambda x: x[3])
+
+#     if isinstance(dataset_ids, list):
+#         obj_list1 = obj_list1[obj_list1['dataset_id'].isin(dataset_ids)].copy()
+
+#     ## Get the keys to the objects that should be removed
+#     obj_list2 = obj_list1.groupby(['dataset_id', 'station_id'])
+
+#     rem_keys = []
+#     for i, row in obj_list2:
+#         rem1 = row.sort_values('KeyDate', ascending=False).iloc[keep_last:]
+#         for i2, row2 in rem1.iterrows():
+#             rem_keys.extend([{'Key': row2['Key'], 'VersionId': row2['VersionId']}])
+
+#     if len(rem_keys) > 0:
+#         ## Split them into 1000 key chunks
+#         rem_keys_chunks = np.array_split(rem_keys, int(np.ceil(len(rem_keys)/1000)))
+
+#         ## Run through and delete the objects...
+#         for keys in rem_keys_chunks:
+#             resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})
+
+#     print(str(len(rem_keys)) + ' objects removed')
+#     print('Updating stations and dataset json files...')
+
+#     dataset_list = obj_list1['dataset_id'].unique().tolist()
 
-#     run_date_dict = {}
 #     for ds in dataset_list:
-#         dataset_id = ds['dataset_id']
+#         ds_stations = put_remote_agg_stations(s3, bucket, ds, threads)
 
-#         ## Get last result dates
-#         try:
-#             key1 = tdm.utils.key_patterns[version]['stations'].format(dataset_id=dataset_id)
-#             obj1 = get_object_s3(key1, connection_config, bucket, counter=0)
-#             rok1 = read_json_zstd(obj1)
-#             if version <= 3:
-#                 rok1b = [r['results_object_key'] if isinstance(r['results_object_key'], dict) else r['results_object_key'][-1] for r in rok1]
-#                 rok2 = pd.DataFrame(rok1b)
-#                 rok2['run_date'] = pd.to_datetime(rok2['run_date'])
-#                 obj_list = rok2.dropna(subset=['run_date'])
-#                 last_run_date = obj_list['run_date'].max()
-#             else:
-#                 rok1b = [r['result_versions'] for r in rok1]
-#                 rok2 = pd.DataFrame(rok1b)
-#                 rok2['version_date'] = pd.to_datetime(rok2['version_date'])
-#                 obj_list = rok2.dropna(subset=['version_date'])
-#                 last_run_date = obj_list['version_date'].max()
-#         except:
-#             last_run_date = run_date1
+#     ### Aggregate all datasets for the bucket
+#     ds_all = put_remote_agg_datasets(s3, bucket, threads)
 
-#         if processing_code in [3]:
-#             last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
-#         elif processing_code in [4, 5]:
-#             last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
-#         elif processing_code in [1, 2, 6]:
-#             if last_run_date < (run_date1 - pd.DateOffset(hours=save_interval_hours) + pd.DateOffset(minutes=5)):
-#                 last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
-#             else:
-#                 last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
-#         else:
-#             raise ValueError('processing_code does not exist.')
-
-#         run_date_dict.update({dataset_id: last_run_date_key})
-
-#     return run_date_dict
+#     return rem_keys
 
 
-def process_results_versions(dataset_list, bucket, s3=None, connection_config=None, public_url=None, version_data=None, system_version=4):
+def process_results_versions(dataset_list, bucket, version_data, s3=None, connection_config=None, public_url=None, system_version=4):
     """
-    Function to process the run date keys for all datasets for the extraction. These are specific to each processing_code.
+    Function to process the run date keys for all datasets for the extraction.
 
     Parameters
     ----------
@@ -885,47 +662,35 @@ def process_results_versions(dataset_list, bucket, s3=None, connection_config=No
     -------
     run_date_dict : dict of str
     """
+    remote1 = dict(s3=s3, connection_config=connection_config, public_url=public_url, bucket=bucket, version=system_version)
+
+    ## Build the ResultVersion model
     mod_date = pd.Timestamp.today(tz='utc').tz_localize(None).round('s')
 
-    if isinstance(version_data, dict):
-        if isinstance(version_data['version_date'], str):
-            version_data['version_date'] = pd.Timestamp(version_data['version_date'])
+    if isinstance(version_data['version_date'], str):
+        version_data['version_date'] = pd.Timestamp(version_data['version_date'])
 
-        version_data['modified_date'] = mod_date
+    version_data['modified_date'] = mod_date
 
-        version_dict = {}
-        for ds in dataset_list:
-            dataset_id = ds['dataset_id']
-            vd = version_data.copy()
-            vd['dataset_id'] = dataset_id
-            version_m = tdm.dataset.ResultVersion(**vd)
-            version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))
+    version_dict = {}
+    for ds in dataset_list:
+        dataset_id = ds['dataset_id']
+        vd = version_data.copy()
+        vd['dataset_id'] = dataset_id
+        version_m = tdm.dataset.ResultVersion(**vd)
+        version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))
 
-    else:
-        version_dict = {'modified_date': mod_date}
-        for ds in dataset_list:
-            dataset_id = ds['dataset_id']
-            vd = {'dataset_id': dataset_id}
+    ## Check to make sure that the passed version date is not earlier than existing versions
+    t1 = Tethys([remote1])
+    datasets = t1._datasets
 
-            ## Get last result dates
-            ## I probably need to make variants for versions 2 and 3
-            if system_version == 4:
-                try:
-                    key1 = tdm.utils.key_patterns[system_version]['results_versions'].format(dataset_id=dataset_id)
-                    obj1 = get_object_s3(key1, bucket, s3, connection_config, public_url, counter=1)
-                    rok1 = read_json_zstd(obj1)
-
-                    version_date = rok1[-1]['version_date']
-                    vd['version_date'] = version_date
-                    version_m = tdm.dataset.ResultVersion(**vd)
-                    version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))
-                except:
-                    run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None).round('s')
-                    vd['version_date'] = run_date1
-                    version_m = tdm.dataset.ResultVersion(**vd)
-                    version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))
-            else:
-                raise NotImplementedError('I need to make variants for versions 2 and 3.')
+    for ds_id, vd in version_dict.items():
+        if ds_id in datasets:
+            new_date = pd.Timestamp(vd['version_date'])
+            vd_old = t1.get_versions(ds_id)
+            old_date = pd.to_datetime([pd.Timestamp(v['version_date']) for v in vd_old]).max()
+            if new_date < old_date:
+                raise ValueError('The new version date is prior to the max old version date. New version date: {new_date}; old max version date: {old_date}. dataset_id: {ds_id}.'.format(new_date=new_date.isoformat(), old_date=old_date.isoformat(), ds_id=ds_id))
 
     return version_dict
 
@@ -1030,24 +795,75 @@ def determine_results_chunks_diffs(source_paths, remote):
     diffs_list = []
     for ds_id, g in paths2.groupby('dataset_id'):
         # print(ds_id)
+
         if ds_id in datasets:
-            _ = tethys.get_versions(ds_id)
+            dataset = datasets[ds_id]
+            method = dataset['method']
+            vd_old = tethys.get_versions(ds_id)
             vd_max = g.version_date.max()
 
+            # Check to make sure new version data is greater or equal to old dates
+            vd0 = [vd['version_date'] for vd in vd_old if pd.Timestamp(vd['version_date']) <= vd_max]
+
+            if not vd0:
+                raise ValueError('The new version date must be greater than or equal to the old version dates.')
+
+            # Determine what version date to use
+            vd1 = [vd['version_date'] for vd in vd_old if pd.Timestamp(vd['version_date']) == vd_max]
+
+            if len(vd1) > 0:
+                stns = tethys.get_stations(ds_id)
+                old_stn_ids = [s['station_id'] for s in stns]
+
+                vd = vd1[0]
+                station_ids = g.station_id.tolist()
+                merge_vals = ['conflict', 'new']
+            else:
+                if method in ['sensor_recording', 'field_activity', 'sample_analysis']:
+                    stns = tethys.get_stations(ds_id)
+                    old_stn_ids = [s['station_id'] for s in stns]
+
+                    vd = vd_old[-1]['version_date']
+                    station_ids = old_stn_ids
+                    merge_vals = ['conflict', 'new', 'old']
+                else:
+                    # vd = vd_max
+                    # station_ids = g.station_id.tolist()
+                    # merge_vals = ['conflict', 'new']
+                    g['key'] = np.nan
+                    g['_merge'] = 'new'
+                    diffs_list.append(g)
+                    continue
+
+            # Determine which stns already exist in Tethys
+            old_stns = list(set([s for s in station_ids if s in old_stn_ids]))
+
+            # Load in the appropriate chunks to check
+            tethys._load_results_chunks(ds_id, old_stns)
+
             chunks1 = []
             for stn_id, c in tethys._results_chunks[ds_id].items():
                 # print(stn_id)
-                chunks2 = chunk_filters(c, version_date=vd_max)
+                chunks2 = chunk_filters(c, version_date=vd)
                 chunks1.extend(chunks2)
 
             chunks_df = pd.DataFrame(chunks1).drop(['version_date', 'content_length', 'height', 'chunk_day'], axis=1).rename(columns={'chunk_hash': 'original_chunk_hash'})
 
+            # Combine to determine which chunks should be assessed
             combo1 = pd.merge(g, chunks_df, on=['dataset_id', 'station_id', 'chunk_id'], indicator=True, how='outer')
-            combo1 = combo1[combo1['_merge'].isin(['both', 'left_only'])]
-            combo1 = combo1[combo1['chunk_hash'] != combo1['original_chunk_hash']].copy()
+            combo1 = combo1.replace({'_merge': {'both': 'conflict', 'left_only': 'new', 'right_only': 'old'}})
+            combo1 = combo1[combo1['_merge'].isin(merge_vals)].copy()
+            combo1['version_date'] = vd_max
 
-            diffs_list.append(combo1.drop(['key', 'original_chunk_hash', '_merge'], axis=1))
+            if 'old' in merge_vals:
+                combo1.loc[combo1['chunk_hash'] == combo1['original_chunk_hash'], '_merge'] = 'new'
+            else:
+                combo1 = combo1[combo1['chunk_hash'] != combo1['original_chunk_hash']].copy()
+
+            diffs_list.append(combo1.drop(['original_chunk_hash'], axis=1))
         else:
+            g['key'] = np.nan
+            g['_merge'] = 'new'
             diffs_list.append(g)
 
     combo2 = pd.concat(diffs_list)
